import nbap.*;
import nbap.test_nbap.*;

/* Universal classes from the OSS runtime library */
import com.oss.asn1.*;
import com.oss.util.*;

/* Java I/O classes */
import java.io.*;

public class ManualDecoding {
    public static void main(String args[]) {
	// Initialize the project
	try {
	    Nbap.initialize();
	} catch (Exception e) {
    	    System.out.println("Initialization exception: " + e);
            System.exit(1);
	}

	Coder coder = Nbap.getPERAlignedCoder();

	// Set the decoder to emit trace output
	coder.enableDecoderDebugging();
	// Set the decoder to leave opentypes undecoded
	coder.disableAutomaticDecoding();

	//------------------------ MANUAL decoding --------------------

	try {
	    // Create the input for the decoder. The decoder will read the
	    // encoded message from the disk file
	    FileInputStream source = new FileInputStream("cellsetuptdd.per");
	    
	    System.out.println();
            System.out.println("The decoder's trace messages ...");
            System.out.println();

            // Decode the outermost PDU
            NBAP_PDU message = (NBAP_PDU)coder.decode(source, new NBAP_PDU());

	    // Determine the CHOICE alternative selected
	    if (message.hasInitiatingMessage()) {
		InitiatingMessage initiatingMessage =
		    (InitiatingMessage)message.getChosenValue();
		completeDecoding(coder, initiatingMessage);
	    } else if (message.hasSuccesfulOutcome()) {
		// For brevity we omit the code to manually decode 
		// opentypes nested in SuccessfulOutcome, UnsuccessfulOutcome
		// and in the Outcome.
		;
	    } else if (message.hasUnsuccesfulOutcome()) {
		// Do manual decoding of opentypes, nested in the
		// UnsuccessulOutcome
		;
	    } else if (message.hasOutcome()) {
		// Do manual decoding of opentypes, nested in the
		// Outcome
		;
	    }
	    
	    // Print the value decoded
	    System.out.println();
	    System.out.println("The NBAP message decoded:");
	    System.out.println(message);
	} catch (Exception e) {
	    System.out.println("Unexpected exception: " + e);
	}
    }
    
    /**
	Decodes opentypes that are nested in the InitiatingMessage.
    */
    public static void completeDecoding(Coder coder, InitiatingMessage message)
	throws DecodeFailedException, DecodeNotSupportedException
    {
	// The InitiatingMessage has the 'value' field that is opentype.
	// nbap.asn specifies that in order to determine the actual type
	// of the 'value' one has to lookup the NBAP-ELEMENTARY-PROCEDURES
	// information object set for the element with '&procedureID' matching
	// the value of the 'procedureID' of this message.
	NBAP_ELEMENTARY_PROCEDURE_OSET procedures = 
	    Test_NBAP.nBAP_ELEMENTARY_PROCEDURES;
	ProcedureID id = message.getProcedureID();
	ASN1Type valueType = null;
	for (int i=0; i<procedures.getSize(); i++) {
	    NBAP_ELEMENTARY_PROCEDURE procedure = 
		(NBAP_ELEMENTARY_PROCEDURE)procedures.getElement(i);
	    if (id.equals(procedure.getProcedureID())) {
		// The matching information object is found.
		// Retrieve the type of the opentype
		valueType = procedure.getInitiatingMessage();
		break;
	    }
	}
	if (valueType != null) {
	    // Decode the nested InitiatingMessage.value opentype
	    OpenType value = message.getValue();
	    InputStream source = value.getEncodedValueAsStream();
	    System.out.println("----------");
	    System.out.println("Decoding the InitiatingMessage.value ...");
	    System.out.println("----------");
	    AbstractData nested = coder.decode(source, valueType);
	    // Now look at the value decoded and decide if it contains more
	    // nested opentypes to decode
	    if (nested instanceof CellSetupRequestTDD) {
		completeDecoding(coder, (CellSetupRequestTDD)nested);
	    } else {
		// More checks for other possible types that can occur
		// in the InitiatingMessage.value
		// ...
		;
	    }
	    value.setDecodedValue(nested);
	}
    }
    
    public static void completeDecoding(Coder coder, CellSetupRequestTDD message)
	throws DecodeFailedException, DecodeNotSupportedException
    {
	// CellSetupRequestTDD has nested opentypes in both the 'protocolIEs'
	// and in the 'protocolExtensions' fields.
	CellSetupRequestTDD.ProtocolIEs protocolIEs = 
	    message.getProtocolIEs();
	int count = protocolIEs.getSize();
	for (int i=0; i<count; i++) {
	    CellSetupRequestTDD.ProtocolIEs.Sequence_ ie = 
		(CellSetupRequestTDD.ProtocolIEs.Sequence_)protocolIEs.getElement(i);
	    completeDecodingOfCellSetupRequestTDD_IE(coder, ie, i);
	}
	if (message.hasProtocolExtensions()) {
	    // Do a similar procedure to decode opentypes, nested in the
	    // 'protocolExtensions'. The corresponding code is omitted for
	    // brevity.
	    ;
	}
    }
    
    public static void completeDecodingOfCellSetupRequestTDD_IE(Coder coder,
	CellSetupRequestTDD.ProtocolIEs.Sequence_ ie, 
	int index) throws DecodeFailedException,
	    DecodeNotSupportedException
    {
	// The 'ie' has a nested opentype in its 'value' field. In order to
	// determine its actual type one has to lookup the 
	// CellSetupRequestTDD-IEs information object set for the element with
	// the '&id' field matching the 'id' field of this IE.
	ProtocolIE_ID id = ie.getId();
	ASN1Type valueType = typeOfProtocolIE(id, 
	    Test_NBAP.cellSetupRequestTDD_IEs);
	if (valueType != null) {
	    // Decode the nested InitiatingMessage.protocolIEs.*.value
	    // opentype
	    OpenType value = ie.getValue();
	    InputStream source = value.getEncodedValueAsStream();
	    System.out.println("----------");
	    System.out.println("Decoding the InitiatingMessage.value.protocolIUe["
		+ index + "].value ...");
	    System.out.println("----------");
	    AbstractData nested = coder.decode(source, valueType);
	    // Now look at the value decoded and decide if it contains more
	    // nested opentypes to decode. ASN.1 definitions in the nbap.asn
	    // indicate that another nested opentype can occur only in the
	    // SCH-Information-Cell-SetupRqstTDD,
	    // PCCPCH-Information-Cell-SetupRqstTDD or in the
	    // TimeSlotConfigurationList-Cell-SetupRqstTDD. For brevity,
	    // we will provide the code that decoded the nested opentypes in the
	    // SCH-Information-Cell-SetupRqstTDD only.
	    if (nested instanceof SCH_Information_Cell_SetupRqstTDD) {
		completeDecoding(coder, 
		    (SCH_Information_Cell_SetupRqstTDD)nested);
	    } else if (nested instanceof PCCPCH_Information_Cell_SetupRqstTDD) {
		// Omitted for brevity.
		;
	    } else if (nested instanceof TimeSlotConfigurationList_Cell_SetupRqstTDD) {
		// Omitted for brevity
		;
	    }
	    value.setDecodedValue(nested);
	}
    }
    
    /**
	Decodes opentypes nested in the SCH-Information-Cell-SetupRqstTDD.
    */
    public static void completeDecoding(Coder coder, 
	SCH_Information_Cell_SetupRqstTDD message) throws DecodeFailedException,
	    DecodeNotSupportedException
    {
	SyncCaseIndicator_Cell_SetupRqstTDD_PSCH syncCaseIndicator =
	    message.getSyncCaseIndicator();
	int count = syncCaseIndicator.getSize();
	for (int i=0; i<count; i++) {
	    SyncCaseIndicator_Cell_SetupRqstTDD_PSCH.Sequence_ ie = 
		(SyncCaseIndicator_Cell_SetupRqstTDD_PSCH.Sequence_)syncCaseIndicator.getElement(i);
	    completeDecodingOfSyncCaseIndicator_IE(coder, ie, i);
	}
	if (message.hasIE_Extensions()) {
	    // Traverse the 'ie-Extensions' list and decode opentypes,
	    // nested in its element. The corresponding code is omitted for
	    // brevity.
	    ;
	}
    }

    public static void completeDecodingOfSyncCaseIndicator_IE(Coder coder,
	SyncCaseIndicator_Cell_SetupRqstTDD_PSCH.Sequence_ ie, 
	int index) throws DecodeFailedException,
	    DecodeNotSupportedException
    {
	// The 'ie' has nested opentype in its 'value' field. In order to
	// determine its actual type one has to lookup the 
	// SyncCaseIndicatorIE-Cell-SetupRqstTDD-PSCH information object set 
	// for the element with the '&id' field matching the 'id' field of 
	// this IE.
	ProtocolIE_ID id = ie.getId();
	ASN1Type valueType = typeOfProtocolIE(id, 
	    Test_NBAP.syncCaseIndicatorIE_Cell_SetupRqstTDD_PSCH);
	if (valueType != null) {
	    // Decode nested opentype
	    OpenType value = ie.getValue();
	    InputStream source = value.getEncodedValueAsStream();
	    System.out.println("----------");
	    System.out.println("Decoding the SCH-Information-Cell-SetupRqstTDD["
		+ index + "].value ...");
	    System.out.println("----------");
	    AbstractData nested = coder.decode(source, valueType);
	    // Now look at the value decoded and decide if it contains more
	    // nested opentypes to decode. ASN.1 definitions in the nbap.asn
	    // indicate that another nested opentype can occur only in the
	    // SyncCaseIndicatorItem-Cell-SetupRqstTDD-PSCH
	    if (nested instanceof SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH) {
		SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH item =
		    (SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH)nested;
		if (item.hasCase1()) {
		    // Decode opentypes nested in the 
		    // Case1Item-Cell-SetupRqstTDD. 
		    Case1_Cell_SetupRqstTDD ies = 
			(Case1_Cell_SetupRqstTDD)item.getChosenValue();
		    int count = ies.getSize();
		    for (int i=0; i<count; i++) {
			Case1_Cell_SetupRqstTDD.Sequence_ ie1 = 
			    (Case1_Cell_SetupRqstTDD.Sequence_)ies.getElement(i);
			completeDecodingOfCase1ItemCell(coder,
			    ie1, i);
		    }
		} else if (item.hasCase2()) {
		    // Decode 'iE-Extensions' in the 
		    // Case2Item-Cell-SetupRqstTDD. The corresponding code is 
		    // omitted for brevity.
		    ;
		}
	    } 
	    value.setDecodedValue(nested);
	}
	
    }
    
    public static void completeDecodingOfCase1ItemCell(Coder coder,
	Case1_Cell_SetupRqstTDD.Sequence_ ie, 
	int index) throws DecodeFailedException,
	    DecodeNotSupportedException
    {
	// The 'ie' has nested opentype in its 'value' field. In order to
	// determine its actual type one has to lookup the 
	// Case1IE-Cell-SetupRqstTDD information object set 
	// for the element with the '&id' field matching the 'id' field of 
	// this IE.
	ProtocolIE_ID id = ie.getId();
	ASN1Type valueType = typeOfProtocolIE(id, 
	    Test_NBAP.case1IE_Cell_SetupRqstTDD);
	if (valueType != null) {
	    // Decode nested opentype
	    OpenType value = ie.getValue();
	    InputStream source = value.getEncodedValueAsStream();
	    System.out.println("----------");
	    System.out.println("Decoding the Case1-Cell-SetupRqstTDD[" +
		+ index + "].value ...");
	    System.out.println("----------");
	    AbstractData nested = coder.decode(source, valueType);
	    // Now look at the value decoded and decide if it contains more
	    // nested opentypes to decode. ASN.1 definitions in the nbap.asn
	    // indicate that another nested opentype can occur only in the
	    // Case1Item-Cell-SetupRqstTDD
	    if (nested instanceof Case1Item_Cell_SetupRqstTDD) {
		Case1Item_Cell_SetupRqstTDD case1Item =
		    (Case1Item_Cell_SetupRqstTDD)nested;
		if (case1Item.hasIE_Extensions()) {
		    // Decode opentypes nested in the 'iE-Extensions'.
		    // The code is omitted for brevity.
		    ;
		}
	    } 
	    value.setDecodedValue(nested);
	}
    }
    
    /**
	Looks up the information object set of the NBAP-PROTOCOL-IES class
	for the information object, when '&id' matches the 'id' parameter. 
	Returns the ASN1Type specified by the matching information object 
	(if any).
    */
    public static ASN1Type typeOfProtocolIE(ProtocolIE_ID id,
	NBAP_PROTOCOL_IES_OSET table)
    {
	int count = table.getSize();
	ASN1Type valueType = null;
	for (int i=0; i<count; i++) {
	    NBAP_PROTOCOL_IES element = (NBAP_PROTOCOL_IES)table.getElement(i);
	    if (id.equals(element.getId())) {
		// The matching element is found. Get the type the opentype
		// is expected to contain.
		return element.getValue();
	    }
	}
	return null;
    }
}

